/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2012, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

//Color Transform With Spatial Filtering Application Main
// main.cpp

#include <stdio.h>
#include <iostream>
#include "transform.h"
#include <stdlib.h>
#include <assert.h>
#include <math.h>

#ifndef WIN32
#include <inttypes.h>

typedef       uint64_t    UInt64;
#else	
typedef       unsigned __int64    UInt64;
#endif

#define Clip3( MinVal, MaxVal, a)   ( ((a)<(MinVal)) ? (MinVal) : (((a)>(MaxVal)) ? (MaxVal) :(a)) )  ///< general min/max clip


int main(int argc, char *argv[])
{
  if(argc<10)
  {
	  printf("Usage:Input Output Width Height nframes BitDepth Format ColorSpace Filter.\n");
	  printf("\tFormat = 1: RGB444 to YCbCr/YCoCg/YFbFr/YFbFr2 420 color space;\n");
	  printf("\tFormat = 2: YCbCr/YCoCg/YFbFr/YFbFr2 420 to RGB444;\n");
	  printf("\tFormat = 3: RGB444 to YCbCr/YCoCg/YFbFr/YFbFr2 444 color space;\n");
	  printf("\tFormat = 4: YCbCr/YCoCg/YFbFr/YFbFr2 444 to RGB444;\n");
	  printf("\tColorSpace = 0: No color transform; 1: YCbCr; 2: YCoCg; 3: YFbFr; 4: YFbFr2.\n");
	  printf("\tFilter = n: No filtering; f: FastVDO; o: SVC; a: Cheap.\n");
	  return -1;
  }
  // Input video details
  Int       m_iSourceWidth;                                   ///< source width in pixel
  Int       m_iSourceHeight;                                  ///< source height in pixel
  Int       m_iFrames;                                       ///< number of frames
  Int       m_iFrameRcvd;                                    ///< number of read frames
  UInt       uiBitDepth;                                     ///< input bit depth
  char*     m_pchInputFile;                                  ///< input file name 
  char*     m_pchOutputFile;                                 ///< output file name


  m_pchInputFile = argv[1];
  m_pchOutputFile = argv[2];
  m_iSourceWidth = atoi(argv[3]);
  m_iSourceHeight = atoi(argv[4]);
  m_iFrames = atoi(argv[5]);
  uiBitDepth = UInt(atoi(argv[6]));

// open input file to read and set corresponding bitdepth
  ColorTransform m_cTVideoIOInputFile;
  m_cTVideoIOInputFile.open( m_pchInputFile, false); 
  m_cTVideoIOInputFile.setbitdepth(uiBitDepth);

  // open output file to write and set corresponding bitdepth

  ColorTransform m_cTVideoIOOutputFile;
  m_cTVideoIOOutputFile.open( m_pchOutputFile, true); 
  m_cTVideoIOOutputFile.setbitdepth( uiBitDepth);

  // allocate buffer for source and destination frame

 // pcPicSrc ->4:4:4 & pcPicDst ->4:2:0

  TComPic*    pcPicSrc = new TComPic;
  TComPic*    pcPicDst = new TComPic;
  pcPicSrc->create( m_iSourceWidth, m_iSourceHeight,1);
  pcPicSrc->setBitDepth(uiBitDepth);
  pcPicDst->create( m_iSourceWidth, m_iSourceHeight,2);
  pcPicDst->setBitDepth(uiBitDepth);
 
   // pcPicSrc2 ->4:2:0 & pcPicDst2 ->4:4:4

  TComPic*    pcPicSrc2 = new TComPic;
  TComPic*    pcPicDst2 = new TComPic;
  pcPicSrc2->create( m_iSourceWidth, m_iSourceHeight,2);
  pcPicSrc2->setBitDepth(uiBitDepth);
  pcPicDst2->create( m_iSourceWidth, m_iSourceHeight,1);
  pcPicDst2->setBitDepth(uiBitDepth);
  
  // Initialization for full file read and write
  Bool  bEos = false;
  m_iFrameRcvd = 0;

  while ( !bEos )
  {
		
		switch(*argv[7])
			{
		    // read one frame of input video file to PIC source buffer
		    // apply Color transform + Spatial Filter (depending on argv[9]) and populate PIC destination buffer
			// write one frame of output video file from PIC destination buffer
            case '1':
				   m_cTVideoIOInputFile.read( pcPicSrc,argv[7]);
				   m_cTVideoIOInputFile.ForTr(pcPicSrc,pcPicDst,argv[8],m_iSourceWidth,m_iSourceHeight,argv[9]);
				   m_cTVideoIOOutputFile.write(pcPicDst,argv[7]);
				   break;

            case '2':
				   m_cTVideoIOInputFile.read( pcPicSrc2,argv[7]);
				   m_cTVideoIOInputFile.InvTr(pcPicSrc2,pcPicDst2,argv[8],m_iSourceWidth,m_iSourceHeight,argv[9]);
				   m_cTVideoIOOutputFile.write(pcPicDst2,argv[7]);
				   break;

		     case '3':
				   m_cTVideoIOInputFile.read( pcPicSrc,argv[7]);
				   m_cTVideoIOInputFile.ForTr(pcPicSrc,pcPicDst2,argv[8],m_iSourceWidth,m_iSourceHeight,argv[9]);
				   m_cTVideoIOOutputFile.write(pcPicDst2,argv[7]);
				   break;

            case '4':
				   m_cTVideoIOInputFile.read( pcPicSrc,argv[7]);
				   m_cTVideoIOInputFile.InvTr(pcPicSrc,pcPicDst2,argv[8],m_iSourceWidth,m_iSourceHeight,argv[9]);
				   m_cTVideoIOOutputFile.write(pcPicDst2,argv[7]);
				   break;

			}

		// increase number of read frames
		m_iFrameRcvd++;
	    
		// check end of file
		bEos = ( m_cTVideoIOInputFile.isEof() == 1 ?   true : false  );
		bEos = ( m_iFrameRcvd == m_iFrames ?    true : bEos   );

  }
 
    pcPicSrc->destroy();
	pcPicDst->destroy();
	pcPicSrc2->destroy();
	pcPicDst2->destroy();
    m_cTVideoIOInputFile.close();
    m_cTVideoIOOutputFile.close();
 
	return 0;
}
