/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2012, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

// Function Definitions
// transform.cpp

#include <stdio.h>
#include <fstream>
#include <iostream>
#include "transform.h"


Void ColorTransform::open( char* pchFile, Bool bWriteMode)
{
  if ( bWriteMode )
  {
    m_cHandle.open( pchFile, ios::binary | ios::out );
    
    if( m_cHandle.fail() )
    {
      printf("\nfailed to write reconstructed YUV file\n");
      exit(0);
    }
  }
  else
  {
    m_cHandle.open( pchFile, ios::binary | ios::in );
    
    if( m_cHandle.fail() )
    {
      printf("\nfailed to open Input YUV file\n");
      exit(0);
    }
  }
  
  return;
}

Void ColorTransform::close()
{
  m_cHandle.close();
}

Bool ColorTransform::isEof()
{
  return m_cHandle.eof();
}

Void ColorTransform::readPlane(Pel* dst, istream& fd, bool is16bit,
                      unsigned int stride,
                      unsigned int width, unsigned int height)
{
  int read_len = width * (is16bit ? 2 : 1);
  unsigned char *buf = new unsigned char[read_len];
  for (UInt y = 0; y < height; y++)
  {
    fd.read(reinterpret_cast<char*>(buf), read_len);
    if (!is16bit) {
      for (UInt x = 0; x < width; x++) {
        dst[x] = buf[x];
      }
    }
    else {
      for (UInt x = 0; x < width; x++) {
        dst[x] = (buf[2*x+1] << 8) | buf[2*x];
      }
    }
    dst += stride;
  }
  delete[] buf;
}

Void ColorTransform::writePlane(ostream& fd, Pel* src, bool is16bit,
                       unsigned int stride,
                       unsigned int width, unsigned int height)
{
  int write_len = width * (is16bit ? 2 : 1);
  unsigned char *buf = new unsigned char[write_len];
  for (UInt y = 0; y < height; y++)
  {
    if (!is16bit) 
    {
      for (UInt x = 0; x < width; x++) 
      {
        buf[x] = (unsigned char) src[x];
      }
    }
    else 
    {
      for (UInt x = 0; x < width; x++) 
      {
        buf[2*x] = src[x] & 0xff;
        buf[2*x+1] = (src[x] >> 8) & 0xff;
      }
    }

    fd.write(reinterpret_cast<char*>(buf), write_len);
    src += stride;
  }
  delete[] buf;
}

Void ColorTransform::read ( TComPic*  rpcPic,char *arg)
{
  // check end-of-file
  if ( isEof() ) return;
  
  Int   iStride = rpcPic->getStride();
  unsigned int width  = rpcPic->getWidth();
  unsigned int height = rpcPic->getHeight();
  bool is16bit = m_fileBitdepth > 8;
  // compute actual YUV width & height
  switch(*arg)
	  {
		  case '1':
			  readPlane(rpcPic->getBuf1(), m_cHandle, is16bit, iStride, width, height);
			  readPlane(rpcPic->getBuf2(), m_cHandle, is16bit, iStride, width, height);
			  readPlane(rpcPic->getBuf3(), m_cHandle, is16bit, iStride, width, height);
		  break;

		  case '2':
			  readPlane(rpcPic->getBuf1(), m_cHandle, is16bit, iStride, width, height);
			  readPlane(rpcPic->getBuf2(), m_cHandle, is16bit, (iStride >> 1), (width >> 1), (height >> 1));
			  readPlane(rpcPic->getBuf3(), m_cHandle, is16bit, (iStride >> 1), (width >> 1), (height >> 1));
		  break;
		  default:
			  readPlane(rpcPic->getBuf1(), m_cHandle, is16bit, iStride, width, height);
			  readPlane(rpcPic->getBuf2(), m_cHandle, is16bit, iStride, width, height);
			  readPlane(rpcPic->getBuf3(), m_cHandle, is16bit, iStride, width, height);
			  break;

	  }
}

Void ColorTransform::write( TComPic* pcPicDst,char *arg)
{
  // compute actual YUV frame size excluding padding size
  Int   iStride = pcPicDst->getStride();
  unsigned int width  = pcPicDst->getWidth();
  unsigned int height = pcPicDst->getHeight();
  bool is16bit = m_fileBitdepth > 8;
  TComPic *dstPic = NULL;
  dstPic = pcPicDst;
  switch(*arg)
	  {
		  case '2':
			  writePlane(m_cHandle, dstPic->getBuf1(), is16bit, iStride, width, height);
			  writePlane(m_cHandle, dstPic->getBuf2(), is16bit, iStride, width, height);
			  writePlane(m_cHandle, dstPic->getBuf3(), is16bit, iStride, width, height); 
		  break;

		  case '1':
			  writePlane(m_cHandle, dstPic->getBuf1(), is16bit, iStride, width, height);
			  writePlane(m_cHandle, dstPic->getBuf2(), is16bit, (iStride >> 1), (width >> 1), (height >> 1));
			  writePlane(m_cHandle, dstPic->getBuf3(), is16bit, (iStride >> 1), (width >> 1), (height >> 1)); 
		  break;
		  default:
			  writePlane(m_cHandle, dstPic->getBuf1(), is16bit, iStride, width, height);
			  writePlane(m_cHandle, dstPic->getBuf2(), is16bit, iStride, width, height);
			  writePlane(m_cHandle, dstPic->getBuf3(), is16bit, iStride, width, height); 
		  break;
	  }

}

void ColorTransform::RGB2YCbCr(TComPic*&  rpcPicOrg, TComPic*&  rpcPicDst)
{
	Int i, iBitDepth;
	Pel *pSrcPlane1, *pSrcPlane2, *pSrcPlane3, *pDstPlane1, *pDstPlane2, *pDstPlane3;
	Int iSize = (rpcPicOrg->getWidth())*(rpcPicOrg->getHeight());
	iBitDepth  = rpcPicOrg->getBitDepth();
	Int OFFSET = 1 << (iBitDepth-1);
	Int MAXVAL = (1 << iBitDepth) -1;
	pSrcPlane1 = rpcPicOrg->getBuf1();
	pSrcPlane2 = rpcPicOrg->getBuf2();
	pSrcPlane3 = rpcPicOrg->getBuf3();
	pDstPlane1 = rpcPicDst->getBuf1();
	pDstPlane2 = rpcPicDst->getBuf2();
	pDstPlane3 = rpcPicDst->getBuf3();

	for(i = 0; i < iSize; i++){
		pDstPlane1[i] =  (Pel)((0.213* pSrcPlane1[i] + 0.715* pSrcPlane2[i] +0.072* pSrcPlane3[i]) + 0.5);
pDstPlane2[i] =  (Pel)((-0.117* pSrcPlane1[i] -0.394* pSrcPlane2[i] + 0.511* pSrcPlane3[i]) + OFFSET);
pDstPlane3[i] =  (Pel)((0.511* pSrcPlane1[i] -0.464* pSrcPlane2[i] -0.047* pSrcPlane3[i]) + OFFSET);
		pDstPlane1[i] = Clip3(0,MAXVAL,pDstPlane1[i]);
		pDstPlane2[i] = Clip3(0,MAXVAL,pDstPlane2[i]);
		pDstPlane3[i] = Clip3(0,MAXVAL,pDstPlane3[i]);
	}
}

void ColorTransform::YCbCr2RGB(TComPic*&  rpcPicOrg, TComPic*&  rpcPicDst)
{
	UInt i, iSize,iBitDepth;
	Pel *pSrcPlane1, *pSrcPlane2, *pSrcPlane3, *pDstPlane1, *pDstPlane2, *pDstPlane3;
	iSize = (rpcPicOrg->getWidth())*(rpcPicOrg->getHeight());
	iBitDepth  = rpcPicOrg->getBitDepth();
	Int OFFSET = 1 << (iBitDepth-1);
	Int MAXVAL = (1 << iBitDepth) -1;
	pSrcPlane1 = rpcPicOrg->getBuf1();
	pSrcPlane2 = rpcPicOrg->getBuf2();
	pSrcPlane3 = rpcPicOrg->getBuf3();
	pDstPlane1 = rpcPicDst->getBuf1();
	pDstPlane2 = rpcPicDst->getBuf2();
	pDstPlane3 = rpcPicDst->getBuf3();

	for(i = 0; i < iSize; i++){
		pDstPlane1[i] = (Pel)(pSrcPlane1[i] + 1.540 *(pSrcPlane3[i] -OFFSET) + 0.5);
pDstPlane2[i] = (Pel)(pSrcPlane1[i] - 0.459 *(pSrcPlane3[i] -OFFSET) - 0.183 *(pSrcPlane2[i] -OFFSET) + 0.5);
pDstPlane3[i] = (Pel)(pSrcPlane1[i] + 1.816 *(pSrcPlane2[i] -OFFSET) + 0.5);
		pDstPlane1[i] = Clip3(0,MAXVAL,pDstPlane1[i]);
		pDstPlane2[i] = Clip3(0,MAXVAL,pDstPlane2[i]);
		pDstPlane3[i] = Clip3(0,MAXVAL,pDstPlane3[i]);

	}
}

void ColorTransform::RGB2YCoCg(TComPic*&  rpcPicOrg, TComPic*&  rpcPicDst)
{
	Int i, iSize,iBitDepth;
	Pel *pSrcPlane1, *pSrcPlane2, *pSrcPlane3, *pDstPlane1, *pDstPlane2, *pDstPlane3, p1, p2, p3;
	iSize = (rpcPicOrg->getWidth())*(rpcPicOrg->getHeight());
	iBitDepth  = rpcPicOrg->getBitDepth();
	Int OFFSET = (1 << (iBitDepth -1)) ;
	Int MAXVAL = (1 << (iBitDepth )) - 1;
	pSrcPlane1 = rpcPicOrg->getBuf1();
	pSrcPlane2 = rpcPicOrg->getBuf2();
	pSrcPlane3 = rpcPicOrg->getBuf3();
	pDstPlane1 = rpcPicDst->getBuf1();
	pDstPlane2 = rpcPicDst->getBuf2();
	pDstPlane3 = rpcPicDst->getBuf3();

	for(i = 0; i < iSize; i++){
		p1 = (*pSrcPlane2++);
		p2 = (*pSrcPlane3++);
		p3 = (*pSrcPlane1++);
		p3 -= p2;
		p2 += (p3 >> 1);
		p1 -= p2;
		p2 += (p1 >> 1);
		p3 = (Pel) p3/2.0;
		p1 = (Pel) p1/2.0;
		p3 += OFFSET;
		p1 += OFFSET;
		p1 = Clip3(0,MAXVAL,p1);
		p2 = Clip3(0,MAXVAL,p2);
		p3 = Clip3(0,MAXVAL,p3);
		*pDstPlane1++ = p2 ;
		*pDstPlane2++ = p3;
		*pDstPlane3++ = p1;
	}
}


void ColorTransform::YCoCg2RGB(TComPic*&  rpcPicOrg, TComPic*&  rpcPicDst)
{
	Int i, iSize,iBitDepth;
	Pel *pSrcPlane1, *pSrcPlane2, *pSrcPlane3, *pDstPlane1, *pDstPlane2, *pDstPlane3, p1, p2, p3;
	iSize = (rpcPicOrg->getWidth())*(rpcPicOrg->getHeight());
	iBitDepth  = rpcPicOrg->getBitDepth();
    Int OFFSET = (1 << (iBitDepth -1)) ;
	Int MAXVAL = (1 << (iBitDepth )) - 1;
	pSrcPlane1 = rpcPicOrg->getBuf1();
	pSrcPlane2 = rpcPicOrg->getBuf2();
	pSrcPlane3 = rpcPicOrg->getBuf3();
	pDstPlane1 = rpcPicDst->getBuf1();
	pDstPlane2 = rpcPicDst->getBuf2();
	pDstPlane3 = rpcPicDst->getBuf3();

	for(i = 0; i < iSize; i++){
		p2 = (*pSrcPlane1++);
		p3 = (*pSrcPlane2++) - OFFSET;
		p3 =p3*2;
		p1 = (*pSrcPlane3++) - OFFSET;		
		p1 = p1*2;
		p2 -= (p1>>1);
		p1 += p2;
		p2 -= (p3>>1);
		p3 += p2;
		p1 = Clip3(0,MAXVAL,p1);
		p2 = Clip3(0,MAXVAL,p2);
		p3 = Clip3(0,MAXVAL,p3);
		*pDstPlane2++ = p1;
		*pDstPlane3++ = p2;
		*pDstPlane1++ = p3;
	}
}

void ColorTransform::RGB2YFbFr(TComPic*&  rpcPicOrg, TComPic*&  rpcPicDst)
{
	Int i, iSize,iBitDepth;
	Pel *pSrcPlane1, *pSrcPlane2, *pSrcPlane3, *pDstPlane1, *pDstPlane2, *pDstPlane3, p1, p2, p3;
	iSize = (rpcPicOrg->getWidth())*(rpcPicOrg->getHeight());
	iBitDepth  = rpcPicOrg->getBitDepth();
	Int OFFSET = (1 << (iBitDepth -1)) ;
	Int MAXVAL = (1 << (iBitDepth )) - 1;
	pSrcPlane1 = rpcPicOrg->getBuf1();
	pSrcPlane2 = rpcPicOrg->getBuf2();
	pSrcPlane3 = rpcPicOrg->getBuf3();
	pDstPlane1 = rpcPicDst->getBuf1();
	pDstPlane2 = rpcPicDst->getBuf2();
	pDstPlane3 = rpcPicDst->getBuf3();

	for(i = 0; i < iSize; i++){
		p3 = pSrcPlane1[i];
		p2 = pSrcPlane3[i];
		p1 = pSrcPlane2[i];
		p3 -= p2;
		p2 += (p3 >> 1);
		p1 -= p2;
		p2 += ((p1*3) >> 3);
		p3 = (Pel) p3/2.0;
		p1 = (Pel) p1/2.0;
		pDstPlane1[i] = Clip3(0,MAXVAL,p2);
		pDstPlane3[i] = Clip3(0,MAXVAL,p3 + OFFSET);
		pDstPlane2[i] = Clip3(0,MAXVAL,p1 + OFFSET);
	}
}


void ColorTransform::YFbFr2RGB(TComPic*&  rpcPicOrg, TComPic*&  rpcPicDst)
{
	Int i, iSize,iBitDepth;
	Pel *pSrcPlane1, *pSrcPlane2, *pSrcPlane3, *pDstPlane1, *pDstPlane2, *pDstPlane3, p1, p2, p3;
	iSize = (rpcPicOrg->getWidth())*(rpcPicOrg->getHeight());
	iBitDepth  = rpcPicOrg->getBitDepth();
    Int OFFSET = (1 << (iBitDepth -1)) ;
	Int MAXVAL = (1 << (iBitDepth )) - 1;
	pSrcPlane1 = rpcPicOrg->getBuf1();
	pSrcPlane2 = rpcPicOrg->getBuf2();
	pSrcPlane3 = rpcPicOrg->getBuf3();
	pDstPlane1 = rpcPicDst->getBuf1();
	pDstPlane2 = rpcPicDst->getBuf2();
	pDstPlane3 = rpcPicDst->getBuf3();
	for(i = 0; i < iSize; i++){
		p1 = (pSrcPlane2[i]) - OFFSET;
		p1 = p1*2;
		p2 = (pSrcPlane1[i]);
		p3 = (pSrcPlane3[i]) - OFFSET;
		p3 =p3*2;
		p2 -= ((p1*3)>>3);
		p1 += p2;
		p2 -= (p3>>1);
		p3 += p2;
		pDstPlane1[i] = Clip3(0,MAXVAL,p3);
		pDstPlane3[i] = Clip3(0,MAXVAL,p2);
		pDstPlane2[i] = Clip3(0,MAXVAL,p1);
	}
}

void ColorTransform::RGB2YFbFrv2(TComPic*&  rpcPicOrg, TComPic*&  rpcPicDst)
{
	Int i, iSize,iBitDepth;
	Pel *pSrcPlane1, *pSrcPlane2, *pSrcPlane3, *pDstPlane1, *pDstPlane2, *pDstPlane3, p1, p2, p3;
	iSize = (rpcPicOrg->getWidth())*(rpcPicOrg->getHeight());
	iBitDepth  = rpcPicOrg->getBitDepth();
	Int OFFSET = (1 << (iBitDepth -1)) ;
	Int MAXVAL = (1 << (iBitDepth )) - 1;
	pSrcPlane1 = rpcPicOrg->getBuf1();
	pSrcPlane2 = rpcPicOrg->getBuf2();
	pSrcPlane3 = rpcPicOrg->getBuf3();
	pDstPlane1 = rpcPicDst->getBuf1();
	pDstPlane2 = rpcPicDst->getBuf2();
	pDstPlane3 = rpcPicDst->getBuf3();

	for(i = 0; i < iSize; i++){
		p3 = pSrcPlane1[i];
		p2 = pSrcPlane3[i];
		p1 = pSrcPlane2[i];
		p3 -= p2;
		p2 += (p3 >> 1);
		p1 -= p2;
		p2 += ((p1*5) >> 3);
		p3 = (Pel) p3/2.0;
		p1 = (Pel) p1/2.0;
		pDstPlane1[i] = Clip3(0,MAXVAL,p2) ;
		pDstPlane3[i] = Clip3(0,MAXVAL,p3 + OFFSET);
		pDstPlane2[i] = Clip3(0,MAXVAL,p1 + OFFSET);
	}
}

void ColorTransform::YFbFr2RGBv2(TComPic*&  rpcPicOrg, TComPic*&  rpcPicDst)
{
	Int i, iSize, iBitDepth;
	Pel *pSrcPlane1, *pSrcPlane2, *pSrcPlane3, *pDstPlane1, *pDstPlane2, *pDstPlane3, p1, p2, p3;
	iSize = (rpcPicOrg->getWidth())*(rpcPicOrg->getHeight());
	iBitDepth  = rpcPicOrg->getBitDepth();
	Int OFFSET = (1 << (iBitDepth -1)) ;
	Int MAXVAL = (1 << (iBitDepth )) - 1;
	pSrcPlane1 = rpcPicOrg->getBuf1();
	pSrcPlane2 = rpcPicOrg->getBuf2();
	pSrcPlane3 = rpcPicOrg->getBuf3();
	pDstPlane1 = rpcPicDst->getBuf1();
	pDstPlane2 = rpcPicDst->getBuf2();
	pDstPlane3 = rpcPicDst->getBuf3();
	for(i = 0; i < iSize; i++){
		p1 = (pSrcPlane2[i]) - OFFSET;
		p1 = p1*2;
		p2 = (pSrcPlane1[i]) ;
		p3 = (pSrcPlane3[i]) - OFFSET;
		p3 =p3*2;
		p2 -= ((p1*5)>>3);
		p1 += p2;
		p2 -= (p3>>1);
		p3 += p2;
		pDstPlane1[i] = Clip3(0,MAXVAL,p3);
		pDstPlane3[i] = Clip3(0,MAXVAL,p2);
		pDstPlane2[i] = Clip3(0,MAXVAL,p1);
	}
}
/*
 * uiLength is the length of the output signal;
 * the length of the input signal is uiLength + uiNTap -1;
 */
void ColorTransform::filtering(Pel * input, Pel * output, int filter[], UInt uiLength, UInt uiNTap, UInt shift,UInt bitdepth)
{
	UInt i,j;
	Int MAXVAL = (1 << (bitdepth+1)) - 1;
	int tmp;
	for( i=0; i < uiLength; i++){
		tmp = 0;
		output[i] = 0;
		for ( j=0; j< uiNTap; j++)
			tmp += input[i+j] * filter[j];
		output[i] = Clip3(0,MAXVAL,(tmp>>shift));
	}

}

void ColorTransform::filtering2D(Pel** sInput, Pel** sOutput, int filter[], UInt uiWidth, UInt uiHeight, UInt uiNTap,UInt bitdepth){
	UInt i,j;
	UInt shift = filter[uiNTap];
	UInt uiPad = (uiNTap-1)>>1;
	short *input, *output;
	input = new short[uiWidth + uiNTap -1];
	output = new short[uiWidth];
	for( i=0; i<uiHeight; i++){
		memcpy(input+uiPad,sInput[i],uiWidth*sizeof(short));
		for ( j=1; j<=uiPad; j++){
			input[uiPad-j] = sInput[i][j];
			input[uiWidth +uiPad -1 +j] = sInput[i][uiWidth-1-j];
		}
		filtering(input, output, filter, uiWidth, uiNTap, shift,bitdepth);
		memcpy(sOutput[i],output,uiWidth*sizeof(short));
	}
	delete[] input;
	delete[] output;

	input = new short[uiHeight + uiNTap -1];
	output = new short[uiHeight];
	for( i=0; i<uiWidth; i++){
		for( j=0; j<uiHeight; j++)
			input[j+uiPad] = sOutput[j][i];
		for ( j=1; j<=uiPad; j++){
			input[uiPad-j] = sOutput[j][i];
			input[uiHeight +uiPad -1 +j] = sOutput[uiHeight-1-j][i];
		}
		filtering(input, output, filter, uiHeight, uiNTap, shift,bitdepth);
		for( j=0; j<uiHeight; j++)
			sOutput[j][i] = output[j];

	}
	delete[] input;
	delete[] output;

}
/*
 * The size of sInput is uiWidth x uiHeight
 * The size of sOutput is (uiWidth >> 1 x (uiHeight >>1)
 */
void ColorTransform::DownSampling(Pel** sInput, Pel** sOutput, int filter[], UInt uiWidth, UInt uiHeight, UInt uiNTap,UInt bitdepth){
	UInt i,j,ii,jj;
	Int MAXVAL =  (1<<(bitdepth)) -1;
	double avg;
	if(uiNTap != 0){
		short ** output;
		output = new short *[uiHeight];
		output[0] = new short[uiHeight*uiWidth];
		for(i=1; i< uiHeight; i++)
			output[i] = output[i-1] + uiWidth;
		filtering2D(sInput, output, filter, uiWidth, uiHeight, uiNTap,bitdepth);
		for(i=0,ii=0;i< (uiHeight>>1);i++,ii+=2)
			for(j=0,jj=0; j<(uiWidth>>1);j++,jj+=2)
				sOutput[i][j] = output[ii][jj];
		delete output[0];
		delete[] output;
	}
	else{
		for(i=0,ii=0;i< (uiHeight>>1);i++,ii+=2)
			for(j=0,jj=0; j<(uiWidth>>1);j++,jj+=2){
				//avg = (sInput[ii][jj] + sInput[ii][jj+1] + sInput[ii+1][jj]+sInput[ii+1][jj+1])/4.0;
				//sOutput[i][j] = Clip3(0,MAXVAL,(Pel) avg) ;//output[ii][jj];
				sOutput[i][j] = sInput[ii][jj];
			}
	}
}
/*
 * The size of sInput is (uiWidth >> 1 x (uiHeight >>1)
 * The size of sOutput is uiWidth x uiHeight
 */
void ColorTransform::UpSampling(Pel** sInput, Pel** sOutput, int filter[], UInt uiWidth, UInt uiHeight, UInt uiNTap,UInt bitdepth){
	UInt i,j,ii,jj;
	if(uiNTap != 0){
		short ** input;
		input = new short *[uiHeight];
		input[0] = new short[uiHeight*uiWidth];
		for(i=1; i< uiHeight; i++)
			input[i] = input[i-1] + uiWidth;
		memset(input[0],0,uiHeight*uiWidth*sizeof(short));
		for(i=0,ii=0;i< (uiHeight>>1);i++,ii+=2){
			for(j=0,jj=0; j<(uiWidth>>1);j++,jj+=2){
				input[ii][jj] = sInput[i][j];
			}

		}
		filtering2D(input, sOutput, filter, uiWidth, uiHeight, uiNTap,bitdepth);
		delete[] input[0]; input[0] = NULL;
		delete[] input; input = NULL;
	}
	else{
		for(i=0,ii=0;i< ((uiHeight>>1)-1);i++,ii+=2)
		{
			for(j=0,jj=0; j<((uiWidth>>1)-1);j++,jj+=2){
				sOutput[ii][jj] = sInput[i][j];
				sOutput[ii+1][jj] = Avg2(sInput[i][j],sInput[i+1][j]);
				sOutput[ii][jj+1] = Avg2(sInput[i][j],sInput[i][j+1]);
				sOutput[ii+1][jj+1] = Avg4(sInput[i][j], sInput[i+1][j],sInput[i][j+1],sInput[i+1][j+1]);
			}
			sOutput[ii][uiWidth-2] = sInput[i][(uiWidth>>1)-1];
			sOutput[ii+1][uiWidth-2] = Avg2(sInput[i][(uiWidth>>1)-1],sInput[i+1][(uiWidth>>1)-1]);
			sOutput[ii][uiWidth-1] = Avg2(sInput[i][(uiWidth>>1)-1],sInput[i][(uiWidth>>1)-2]);
			sOutput[ii+1][uiWidth-1] = Avg4(sInput[i][(uiWidth>>1)-1], sInput[i+1][(uiWidth>>1)-1],sInput[i][(uiWidth>>1)-2],sInput[i+1][(uiWidth>>1)-2]);
		}
		for(j=0,jj=0; j<(uiWidth>>1)-1;j++,jj+=2){
			sOutput[uiHeight-2][jj] = sInput[(uiHeight>>1)-1][j];
			sOutput[uiHeight-1][jj] = Avg2(sInput[(uiHeight>>1)-1][j],sInput[(uiHeight>>1)-2][j]);
			sOutput[uiHeight-2][jj+1] = Avg2(sInput[(uiHeight>>1)-1][j],sInput[(uiHeight>>1)-1][j+1]);
			sOutput[uiHeight-1][jj+1] = Avg4(sInput[(uiHeight>>1)-1][j], sInput[(uiHeight>>1)-2][j],sInput[(uiHeight>>1)-1][j+1],sInput[(uiHeight>>1)-1-2][j+1]);
		}
		sOutput[uiHeight-2][uiWidth-2] = sInput[(uiHeight>>1)-1][(uiWidth>>1)-1];
		sOutput[uiHeight-1][uiWidth-2] = Avg2(sInput[(uiHeight>>1)-1][(uiWidth>>1)-1],sInput[(uiHeight>>1)-2][(uiWidth>>1)-1]);
		sOutput[uiHeight-2][uiWidth-1] = Avg2(sInput[(uiHeight>>1)-1][(uiWidth>>1)-1],sInput[(uiHeight>>1)-1][(uiWidth>>1)-2]);
		sOutput[uiHeight-1][uiWidth-1] = Avg4(sInput[(uiHeight>>1)-1][(uiWidth>>1)-1], sInput[(uiHeight>>1)-2][(uiWidth>>1)-1],sInput[(uiHeight>>1)-1][(uiWidth>>1)-2],sInput[(uiHeight>>1)-2][(uiWidth>>1)-2]);

	}

}


Void ColorTransform::setbitdepth(UInt bitdepth)
{
	m_fileBitdepth = bitdepth;
}

Void ColorTransform::ForTr(TComPic*&  rpcPicOrg, TComPic*&  rpcPicDst,char *arg1,Int m_iSourceWidth,Int m_iSourceHeight,char *arg2)
{
    TComPic*    pcPicIntDst = new TComPic;
	pcPicIntDst->create( m_iSourceWidth, m_iSourceHeight,1);
	pcPicIntDst->setBitDepth(m_fileBitdepth);

	Pel **Buf0= (Pel**)malloc(m_iSourceHeight * sizeof(Pel*));
    Pel **Buf1= (Pel**)malloc(m_iSourceHeight * sizeof(Pel*));
	for (int i = 0; i < m_iSourceHeight; i++) 
		  {
			Buf0[i] = (Pel*)malloc(m_iSourceWidth * sizeof(Pel));
			Buf1[i] = (Pel*)malloc(m_iSourceWidth * sizeof(Pel));
		  }

    Pel **InterBuf1= (Pel**)malloc((m_iSourceHeight >> 1) * sizeof(Pel*));
    Pel **InterBuf2= (Pel**)malloc((m_iSourceHeight >> 1) * sizeof(Pel*));
    for (int i = 0; i < (m_iSourceHeight >> 1); i++) 
		  {
			InterBuf1[i] = (Pel*)malloc((m_iSourceWidth >> 1) * sizeof(Pel));
			InterBuf2[i] = (Pel*)malloc((m_iSourceWidth >> 1) * sizeof(Pel));
		  }

	switch(*arg1)
			{

				case '1':
					   RGB2YCbCr(rpcPicOrg,pcPicIntDst);
					   break;
				case '2':
					   RGB2YCoCg(rpcPicOrg,pcPicIntDst);
					  // m_fileBitdepth +=1;
					   break;
				case '3':
					   RGB2YFbFr(rpcPicOrg,pcPicIntDst);
					  // m_fileBitdepth +=1;
					  break;
				case '4':
					   RGB2YFbFrv2(rpcPicOrg,pcPicIntDst);
					  // m_fileBitdepth +=1;
                      break;
				default:
					::memcpy ( pcPicIntDst->getBuf1(), rpcPicOrg->getBuf1(), sizeof (Pel) * ( m_iSourceWidth) * ( m_iSourceHeight ) );
					::memcpy ( pcPicIntDst->getBuf2(), rpcPicOrg->getBuf2(), sizeof (Pel) * ( m_iSourceWidth) * ( m_iSourceHeight ) );
					::memcpy ( pcPicIntDst->getBuf3(), rpcPicOrg->getBuf3(), sizeof (Pel) * ( m_iSourceWidth) * ( m_iSourceHeight ) );
					break;
			}

        Pel * buf2 = pcPicIntDst->getBuf2();
		Pel * buf3 = pcPicIntDst->getBuf3();
		for(int i=0; i < m_iSourceHeight;i++)
			{
				memcpy(Buf0[i],buf2,sizeof(Pel)*m_iSourceWidth);
				memcpy(Buf1[i],buf3,sizeof(Pel)*m_iSourceWidth);
				buf2+=m_iSourceWidth;
				buf3+=m_iSourceWidth;
			}
	::memcpy ( rpcPicDst->getBuf1(), pcPicIntDst->getBuf1(), sizeof (Pel) * ( m_iSourceWidth) * ( m_iSourceHeight ) );
    Pel * buf6 = rpcPicDst->getBuf2();
	Pel * buf7 = rpcPicDst->getBuf3();
    switch (*arg2)
			{
				case 'f':
					{
						int dfilter[12] = {1,0,-3,0,10,16,10,0,-3,0,1,5}; // 5 
						DownSampling(Buf0, InterBuf1,dfilter,m_iSourceWidth,m_iSourceHeight,11,m_fileBitdepth);
						DownSampling(Buf1, InterBuf2,dfilter,m_iSourceWidth,m_iSourceHeight,11,m_fileBitdepth);
						for(int i=0; i < (m_iSourceHeight >> 1);i++)
									{
										memcpy(buf6,InterBuf1[i],sizeof(Pel)*(m_iSourceWidth >> 1));
										memcpy(buf7,InterBuf2[i],sizeof(Pel)*(m_iSourceWidth >> 1));
										buf6+=(m_iSourceWidth >> 1);
										buf7+=(m_iSourceWidth >> 1);
									}
						break;
					}
				case 'o' :
					{
						int dfilter[14] = {2,0,-4,-3,5,19,26,19,5,-3,-4,0,2,6}; //6
						DownSampling(Buf0, InterBuf1,dfilter,m_iSourceWidth,m_iSourceHeight,13,m_fileBitdepth);
						DownSampling(Buf1, InterBuf2,dfilter,m_iSourceWidth,m_iSourceHeight,13,m_fileBitdepth);
						for(int i=0; i < (m_iSourceHeight >> 1);i++)
									{
										memcpy(buf6,InterBuf1[i],sizeof(Pel)*(m_iSourceWidth >> 1));
										memcpy(buf7,InterBuf2[i],sizeof(Pel)*(m_iSourceWidth >> 1));
										buf6+=(m_iSourceWidth >> 1);
										buf7+=(m_iSourceWidth >> 1);
									}
						break;
					}
				case 'a' :
					{
						int dfilter[14] = {2,0,-4,-3,5,19,26,19,5,-3,-4,0,2,6}; //6
						DownSampling(Buf0, InterBuf1,dfilter,m_iSourceWidth,m_iSourceHeight,0,m_fileBitdepth);
						DownSampling(Buf1, InterBuf2,dfilter,m_iSourceWidth,m_iSourceHeight,0,m_fileBitdepth);
						for(int i=0; i < (m_iSourceHeight >> 1);i++)
									{
										memcpy(buf6,InterBuf1[i],sizeof(Pel)*(m_iSourceWidth >> 1));
										memcpy(buf7,InterBuf2[i],sizeof(Pel)*(m_iSourceWidth >> 1));
										buf6+=(m_iSourceWidth >> 1);
										buf7+=(m_iSourceWidth >> 1);
									}
						break;
					}

				default:
					::memcpy ( rpcPicDst->getBuf2(), pcPicIntDst->getBuf2(), sizeof (Pel) * ( m_iSourceWidth) * ( m_iSourceHeight ) );
					::memcpy ( rpcPicDst->getBuf3(), pcPicIntDst->getBuf3(), sizeof (Pel) * ( m_iSourceWidth) * ( m_iSourceHeight ) );
					break;
			}
	for (int i = 0; i < m_iSourceHeight; i++) 
		if(Buf0[i]) { free(Buf0[i]);Buf0[i] = NULL;}
	for (int i = 0; i < m_iSourceHeight; i++) 
		if(Buf1[i]) { free(Buf1[i]); Buf1[i] = NULL;}
    for (int i = 0; i < (m_iSourceHeight>>1); i++) 
		if(InterBuf1[i]) {free(InterBuf1[i]); InterBuf1[i] = NULL;}
			
	for (int i = 0; i < (m_iSourceHeight>>1); i++) 
		if(InterBuf2[i]) {	free(InterBuf2[i]); InterBuf2[i] = NULL;}
	
	if(Buf0) {free(Buf0); Buf0 = NULL;}
	if(Buf1) {free(Buf1); Buf1 = NULL;}
	if(InterBuf1) {free(InterBuf1); InterBuf1 = NULL;}
	if(InterBuf2) {free(InterBuf2); InterBuf2 = NULL;}
	pcPicIntDst->destroy();
	delete pcPicIntDst;
	pcPicIntDst = NULL;}

Void ColorTransform::InvTr(TComPic*&  rpcPicOrg, TComPic*&  rpcPicDst,char *arg1,Int m_iSourceWidth,Int m_iSourceHeight,char *arg2)
{
	TComPic*    pcPicIntDst = new TComPic;
	pcPicIntDst->create( m_iSourceWidth, m_iSourceHeight,1);
	pcPicIntDst->setBitDepth( m_fileBitdepth);

	::memcpy ( pcPicIntDst->getBuf1(), rpcPicOrg->getBuf1(), sizeof (Pel) * ( m_iSourceWidth) * ( m_iSourceHeight ) );

    Pel **Buf0= (Pel**)malloc((m_iSourceHeight >> 1) * sizeof(Pel*));
    Pel **Buf1= (Pel**)malloc((m_iSourceHeight >> 1) * sizeof(Pel*));
    for (int i = 0; i < (m_iSourceHeight >> 1); i++) 
		  {
			Buf0[i] = (Pel*)malloc((m_iSourceWidth >> 1) * sizeof(Pel));
			Buf1[i] = (Pel*)malloc((m_iSourceWidth >> 1) * sizeof(Pel));
		  }

	Pel * buf2 = rpcPicOrg->getBuf2();
	Pel * buf3 = rpcPicOrg->getBuf3();
	for(int i=0; i < (m_iSourceHeight >> 1);i++)
		{
		memcpy(Buf0[i],buf2,sizeof(Pel)*(m_iSourceWidth >> 1));
		memcpy(Buf1[i],buf3,sizeof(Pel)*(m_iSourceWidth >> 1));
		buf2+=(m_iSourceWidth >> 1);
		buf3+=(m_iSourceWidth >> 1);
		}

	Pel **Buf4= (Pel**)malloc(m_iSourceHeight * sizeof(Pel*));
    Pel **Buf5= (Pel**)malloc(m_iSourceHeight * sizeof(Pel*));
	for (int i = 0; i < m_iSourceHeight; i++) 
		  {
			Buf4[i] = (Pel*)malloc(m_iSourceWidth * sizeof(Pel));
			Buf5[i] = (Pel*)malloc(m_iSourceWidth * sizeof(Pel));
		  }
   Pel * buf6 = pcPicIntDst->getBuf2();
   Pel * buf7 = pcPicIntDst->getBuf3();

	switch (*arg2)
		{
			case 'f':
					{
						int ufilter[8] = {-1,0,5,8,5,0,-1,3}; //3		
						//int ufilter[12] = {1,0,-3,0,10,16,10,0,-3,0,1,4}; //4	
						UpSampling(Buf0,Buf4 ,ufilter,m_iSourceWidth ,m_iSourceHeight ,7,m_fileBitdepth);
						UpSampling(Buf1,Buf5 ,ufilter,m_iSourceWidth ,m_iSourceHeight ,7,m_fileBitdepth);
						for(int i=0; i < m_iSourceHeight;i++)
								{
									memcpy(buf6,Buf4[i],sizeof(Pel)*m_iSourceWidth);
									memcpy(buf7,Buf5[i],sizeof(Pel)*m_iSourceWidth);
									buf6+=m_iSourceWidth;
									buf7+=m_iSourceWidth;
								}
						break;
					}
			case 'o' :
					{
						int ufilter[8] = {-3,0,19,32,19,0,-3,5};			
						UpSampling(Buf0,Buf4 ,ufilter,m_iSourceWidth ,m_iSourceHeight ,7,m_fileBitdepth);
						UpSampling(Buf1,Buf5 ,ufilter,m_iSourceWidth ,m_iSourceHeight ,7,m_fileBitdepth);
						for(int i=0; i < m_iSourceHeight;i++)
								{
									memcpy(buf6,Buf4[i],sizeof(Pel)*m_iSourceWidth);
									memcpy(buf7,Buf5[i],sizeof(Pel)*m_iSourceWidth);
									buf6+=m_iSourceWidth;
									buf7+=m_iSourceWidth;
								}
						break;
					}
			case 'a':
					{
						int ufilter[4] = {1,2,1,1};			
						UpSampling(Buf0,Buf4 ,ufilter,m_iSourceWidth ,m_iSourceHeight ,3,m_fileBitdepth);
						UpSampling(Buf1,Buf5 ,ufilter,m_iSourceWidth ,m_iSourceHeight ,3,m_fileBitdepth);
						for(int i=0; i < m_iSourceHeight;i++)
								{
									memcpy(buf6,Buf4[i],sizeof(Pel)*m_iSourceWidth);
									memcpy(buf7,Buf5[i],sizeof(Pel)*m_iSourceWidth);
									buf6+=m_iSourceWidth;
									buf7+=m_iSourceWidth;
								}
						break;
					}

			default:
					//::memcpy ( pcPicIntDst->getBuf1(), rpcPicOrg->getBuf1(), sizeof (Pel) * ( m_iSourceWidth) * ( m_iSourceHeight ) );
					::memcpy ( pcPicIntDst->getBuf2(), rpcPicOrg->getBuf2(), sizeof (Pel) * ( m_iSourceWidth) * ( m_iSourceHeight ) );
					::memcpy ( pcPicIntDst->getBuf3(), rpcPicOrg->getBuf3(), sizeof (Pel) * ( m_iSourceWidth) * ( m_iSourceHeight ) );

				break;
		}

	switch(*arg1)
	   {
		   case '1':
			   YCbCr2RGB(pcPicIntDst,rpcPicDst);
			   break;
		   case '2':
			   YCoCg2RGB(pcPicIntDst,rpcPicDst);
			   break;
		   case '3':   
			   YFbFr2RGB(pcPicIntDst,rpcPicDst);
			   break;
		   case '4':   
			   YFbFr2RGBv2(pcPicIntDst,rpcPicDst);
			   break;
		   default:
				     ::memcpy ( rpcPicDst->getBuf1(), pcPicIntDst->getBuf1(), sizeof (Pel) * ( m_iSourceWidth) * ( m_iSourceHeight ) );
				     ::memcpy ( rpcPicDst->getBuf2(), pcPicIntDst->getBuf2(), sizeof (Pel) * ( m_iSourceWidth) * ( m_iSourceHeight ) );
					 ::memcpy ( rpcPicDst->getBuf3(), pcPicIntDst->getBuf3(), sizeof (Pel) * ( m_iSourceWidth) * ( m_iSourceHeight ) );
			   break;
	   }

	for (int i = 0; i < (m_iSourceHeight>>1); i++) 
		if(Buf0[i]) {free(Buf0[i]); Buf0[i] = NULL;}
		
			
	for (int i = 0; i < (m_iSourceHeight>>1); i++) 
		if(Buf1[i]) {free(Buf1[i]); Buf1[i] = NULL;}
	for (int i = 0; i < (m_iSourceHeight); i++)
		if(Buf4[i]){free(Buf4[i]);Buf4[i] = NULL;}
			 
	for (int i = 0; i < (m_iSourceHeight); i++) 
		if(Buf5[i]){free(Buf5[i]); Buf5[i] = NULL;}
			
	if(Buf0){free(Buf0); Buf0 = NULL;}
	if(Buf1){free(Buf1); Buf1 = NULL;}
	if(Buf4){free(Buf4); Buf4 = NULL;}
	if(Buf5){free(Buf5); Buf5 = NULL;}
	pcPicIntDst->destroy();
	delete pcPicIntDst;
	pcPicIntDst = NULL;
}

Void TComPic::create( Int iPicWidth, Int iPicHeight,char arg)
{
  m_iPicWidth       = iPicWidth;
  m_iPicHeight      = iPicHeight;
  
  if(arg==1)
  {
		  m_apiPicBuf1      = (Pel*)xMalloc( Pel, ( m_iPicWidth) * ( m_iPicHeight ));
		  m_apiPicBuf2      = (Pel*)xMalloc( Pel, ( m_iPicWidth) * ( m_iPicHeight ));
		  m_apiPicBuf3      = (Pel*)xMalloc( Pel, ( m_iPicWidth) * ( m_iPicHeight ));
  }

  else if (arg==2)
  {
		  m_apiPicBuf1      = (Pel*)xMalloc( Pel, ( m_iPicWidth) * ( m_iPicHeight ));
		  m_apiPicBuf2      = (Pel*)xMalloc( Pel, ( m_iPicWidth >> 1) * ( m_iPicHeight >> 1 ));
		  m_apiPicBuf3      = (Pel*)xMalloc( Pel, ( m_iPicWidth >> 1) * ( m_iPicHeight >> 1 ));
  }
  
  return;
}

Void TComPic::destroy()
{
  if( m_apiPicBuf1 ){ xFree( m_apiPicBuf1 );    m_apiPicBuf1 = NULL; }
  if( m_apiPicBuf2 ){ xFree( m_apiPicBuf2 );    m_apiPicBuf2 = NULL; }
  if( m_apiPicBuf3 ){ xFree( m_apiPicBuf3 );    m_apiPicBuf3 = NULL; }
}

Void  TComPic::copyToPic (TComPic*  pcPicDst)
{
  assert( m_iPicWidth  == pcPicDst->getWidth()  );
  assert( m_iPicHeight == pcPicDst->getHeight() );
  
  ::memcpy ( pcPicDst->getBuf1(), m_apiPicBuf1, sizeof (Pel) * ( m_iPicWidth ) * ( m_iPicHeight ) );
  ::memcpy ( pcPicDst->getBuf2(), m_apiPicBuf2, sizeof (Pel) * ( m_iPicWidth ) * ( m_iPicHeight ) );
  ::memcpy ( pcPicDst->getBuf3(), m_apiPicBuf3, sizeof (Pel) * ( m_iPicWidth ) * ( m_iPicHeight ) );
  return;
}

