/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2012, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

//Function declerations
// transform.h

#ifndef __FORWARD_COLOR_TRANSFORM__
#define __FORWARD_COLOR_TRANSFORM__

#include <stdio.h>
#include <fstream>
#include <iostream>
#include <assert.h>

#ifdef WIN32
#define xMalloc( type, len )        _aligned_malloc( sizeof(type)*(len), 32 )
#define xFree( ptr )                _aligned_free  ( ptr )
#else
#define xMalloc( type, len )        malloc( sizeof(type)*(len) )
#define xFree( ptr )                free  ( ptr )
#endif
#define Clip3( MinVal, MaxVal, a)   ( ((a)<(MinVal)) ? (MinVal) : (((a)>(MaxVal)) ? (MaxVal) :(a)) )  ///< general min/max clip
#define Avg2(a,b) (Pel)((a+b)/2.0)
#define Avg4(a,b, c, d) (Pel)((a+b+c+d)/4.0)
typedef void Void;
typedef unsigned int UInt;
typedef int Int;
typedef bool Bool;
typedef short Pel;

using namespace std;

// ====================================================================================================================
// Class definition
// ====================================================================================================================
class TComPic
{
private:
  
  // ------------------------------------------------------------------------------------------------
  //   buffer
  // ------------------------------------------------------------------------------------------------
  
  Pel*  m_apiPicBuf1;            
  Pel*  m_apiPicBuf2;
  Pel*  m_apiPicBuf3;
  
  // ------------------------------------------------------------------------------------------------
  //  Parameter for general YUV buffer usage
  // ------------------------------------------------------------------------------------------------
  
  Int   m_iPicWidth;            ///< Width of picture
  Int   m_iPicHeight;           ///< Height of picture
  UInt   m_iPicBitDepth;
public:
	TComPic()         {}
	virtual ~TComPic()  {}
  
  // ------------------------------------------------------------------------------------------------
  //  Memory management
  // ------------------------------------------------------------------------------------------------
  
  Void  create      ( Int iPicWidth, Int iPicHeight,char arg);
  Void  destroy     ();
  
  // ------------------------------------------------------------------------------------------------
  //  Get information of picture
  // ------------------------------------------------------------------------------------------------
  
  Int   getWidth    ()     { return  m_iPicWidth;    }
  Int   getHeight   ()     { return  m_iPicHeight;   }
  Int   getStride   ()     { return m_iPicWidth; }
  Int   getBitDepth ()     { return m_iPicBitDepth;}
  void   setBitDepth (UInt uiBitDepth)      {m_iPicBitDepth = uiBitDepth;}
  
  // ------------------------------------------------------------------------------------------------
  //  Access function for picture buffer
  // ------------------------------------------------------------------------------------------------
  
  //  Access starting position of original picture
  Pel*  getBuf1 ()     { return  m_apiPicBuf1;    }
  Pel*  getBuf2  ()     { return  m_apiPicBuf2;    }
  Pel*  getBuf3   ()     { return  m_apiPicBuf3;    }

  // copy to pivture function
  Void  copyToPic       ( TComPic*  pcPicDst );
};// END CLASS DEFINITION 

class ColorTransform
{
private:
  fstream   m_cHandle;                                      ///< file handle
  UInt m_fileBitdepth; ///< bitdepth of input/output video file
  Int m_bitdepthShift;  ///< number of bits to increase or decrease image by before/after write/read
  
public:
	ColorTransform()           {}
	virtual ~ColorTransform()  {}
  
  Void  setbitdepth(UInt);
  Void  open  ( char* pchFile, Bool bWriteMode); ///< open or create file
  Void  close ();                                           ///< close file

  Void  read  ( TComPic* pPelSrc,char *arg );								///< read  one YUV frame 
  Void  write ( TComPic* pPelDst,char *arg );								///< write one YUV frame 

  Void readPlane(Pel* dst, istream& fd, bool is16bit,
                      unsigned int stride,
                      unsigned int width, unsigned int height);

  Void writePlane(ostream& fd, Pel* src, bool is16bit,
                       unsigned int stride,
                       unsigned int width, unsigned int height);
  
  Bool  isEof ();                                           ///< check end-of-file
 
  void RGB2YCbCr(TComPic*&  rpcPicOrg, TComPic*&  rpcPicDst);
  void YCbCr2RGB(TComPic*&  rpcPicOrg, TComPic*&  rpcPicDst);
  void RGB2YCoCg(TComPic*&  rpcPicOrg, TComPic*&  rpcPicDst);
  void YCoCg2RGB(TComPic*&  rpcPicOrg, TComPic*&  rpcPicDst);
  void RGB2YFbFr(TComPic*&  rpcPicOrg, TComPic*&  rpcPicDst);
  void YFbFr2RGB(TComPic*&  rpcPicOrg, TComPic*&  rpcPicDst);
  void RGB2YFbFrv2(TComPic*&  rpcPicOrg, TComPic*&  rpcPicDst);
  void YFbFr2RGBv2(TComPic*&  rpcPicOrg, TComPic*&  rpcPicDst);
  void filtering(short * input, short * output, int filter[], UInt uiLength, UInt uiNTap, UInt shift,UInt bitdepth);
  void filtering2D(short** sInput, short** sOutput, int filter[], UInt uiWidth, UInt uiHeight, UInt uiNTap,UInt bitdepth);
  void DownSampling(short** sInput, short** sOutput, int filter[], UInt uiWidth, UInt uiHeight, UInt uiNTap,UInt bitdepth);
  void UpSampling(short** sInput, short** sOutput, int filter[], UInt uiWidth, UInt uiHeight, UInt uiNTap,UInt bitdepth);
  void ForTr(TComPic*&  rpcPicOrg, TComPic*&  rpcPicDst,char *arg1,Int m,Int n,char *arg2);
  void InvTr(TComPic*&  rpcPicOrg, TComPic*&  rpcPicDst,char *arg1,Int m,Int n,char *arg2);
};

#endif // __FORWARD_COLOR_TRANSFORM__

