#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <inttypes.h>
#define uint8 unsigned char
#define uint16 unsigned short

void PsnrOneByte(uint8 *b1, uint8 * b2, int w, int h,double *psnr);
void PsnrTwoBytes(uint16 *b1, uint16 * b2, int w, int h, int MAX,double *psnr);
int main(int argc, char* argv[])
{
	if(argc<6)
	{
		printf("Usage: width, height, bitdepth, nframes,file1, file2.\n");
		return -1;
	}
	int w = atoi(argv[1]);
	int h = atoi(argv[2]);
	int size = w*h*3;
	FILE * F1, *F2;;
	if( (F1 = fopen(argv[5],"r")) == NULL)
		return -2;
	if( (F2 = fopen(argv[6],"r")) == NULL)
		return -3;
	int bitdepth = atoi(argv[3]);
	int nframes = atoi(argv[4]);
	if(nframes <= 0)
		return -4;
	//double apsnr[nframes][3];
	double addpsnr[3];
	double psnr[3];
	addpsnr[0] = 0.0;
	addpsnr[1] = 0.0;
	addpsnr[2] = 0.0;
	long offset = 0;
	if (bitdepth==8)
	{
		uint8 * b1, *b2;
		b1 = (uint8*) malloc(size*sizeof(uint8));
		b2 = (uint8*) malloc(size*sizeof(uint8));
		for(int i = 0; i < nframes; i++)
		{
			fseek(F1,offset,SEEK_SET);
			fseek(F2,offset,SEEK_SET);
			offset += size*sizeof(uint16);
			fread(b1,sizeof(uint8),size,F1);
			 fread(b2,sizeof(uint8),size,F2);
			 PsnrOneByte(b1,b2,w,h,psnr);
			 addpsnr[0]+=psnr[0];
			 addpsnr[1]+=psnr[1];
			 addpsnr[2]+=psnr[2];
			 //apsnr[i][1]=psnr[1];
			 //apsnr[i][2]=psnr[2];
			 //apsnr[i][0]=psnr[0];
		}
		free(b1); b1 = NULL;
		free(b2); b2 = NULL;
	}
	else if ((bitdepth > 9) && (bitdepth<16))
	{
		
		uint16 * b1, *b2;
		b1 = (uint16*) malloc(size*sizeof(uint16));
		b2 = (uint16*) malloc(size*sizeof(uint16));

		
		for(int i = 0; i < nframes; i++)
		{
			
			fseek(F1,offset,SEEK_SET);
			fseek(F2,offset,SEEK_SET);
			offset += size*sizeof(uint16);
			 fread(b1,sizeof(uint16),size,F1);
			 fread(b2,sizeof(uint16),size,F2);
			 PsnrTwoBytes(b1,b2,w,h,(255<<(bitdepth-8)),psnr);
			 addpsnr[0]+=psnr[0];
			 addpsnr[1]+=psnr[1];
			 addpsnr[2]+=psnr[2];
			 //apsnr[i][1]=psnr[1];
			 //apsnr[i][2]=psnr[2];
			 //apsnr[i][0]=psnr[0];
		}
		addpsnr[0]=addpsnr[0]/nframes;
		addpsnr[1]=addpsnr[1]/nframes;
		addpsnr[2]=addpsnr[2]/nframes;
		printf("Total frames: %6d %8.4lf %8.4lf %8.4lf\n",nframes,addpsnr[0],psnr[1],psnr[2]);
		free(b1); b1 = NULL;
		free(b2); b2 = NULL;
	}
}

void PsnrOneByte(uint8 *b1, uint8 * b2, int w, int h, double *psnr)
{
	double ssd1, ssd2, ssd3;
	double psnr1,psnr2,psnr3;
	ssd1 = 0.0;
	ssd2 =0.0;
	ssd3 = 0.0;
	uint8 * tmp1 = b1;
	uint8 * tmp2 = b2;
	int size = w*h;
	int skip1 = size;
	int skip2 = size+size;
	double fRefValue1 = 255*255*size;
	int i, i2,i3;
	for(i=0, i2=size, i3 = (size+size);i<size;i++, i2++,i3++)
	{
		int diff = ((int)b1[i] - (int)b2[i]);
		ssd1 += diff*diff;
		diff = ((int)b1[i2] - (int)b2[i2]);
		ssd2 += diff*diff;
		diff = ((int)b1[i3] - (int)b2[i3]);
		ssd3 += diff*diff;
	}

	psnr[0] = (ssd1 ? 10.0 * log10( fRefValue1 / ssd1) : 99.99 );
	psnr[1] = (ssd2 ? 10.0 * log10( fRefValue1 / ssd2) : 99.99 );
	psnr[2] = (ssd3 ? 10.0 * log10( fRefValue1 / ssd3) : 99.99 );

	printf("%8.4lf %8.4lf %8.4lf\n",psnr[0],psnr[1],psnr[2]);
}

void PsnrTwoBytes(uint16 *b1, uint16 * b2, int w, int h, int MAX,double *psnr)
{
	uint64_t ssd1, ssd2, ssd3;
	double psnr1,psnr2,psnr3;
	ssd1 = 0.0;
	ssd2 =0.0;
	ssd3 = 0.0;
	uint16 * tmp1 = b1;
	uint16 * tmp2 = b2;
	int size = w*h;
	int skip1 = size;
	int skip2 = size+size;
	double fRefValue1 = (double)MAX*MAX*size;
	int i, i2,i3;
	for(i=0, i2=size, i3 = (size+size);i<size;i++, i2++,i3++)
	{
	  //int diff = ((int)b1[i] - (int)b2[i]);
	  int diff = (int)(b1[i]-b2[i]);
		ssd1 += diff*diff;
		diff = (int)(b1[i2] - b2[i2]);
		ssd2 += diff*diff;
			diff = (int)(b1[i3] - b2[i3]);
		ssd3 += diff*diff;
	}

	psnr[0] = (ssd1 ? 10.0 * log10( fRefValue1 / (double)ssd1) : 99.99 );
		   psnr[1] = (ssd2 ? 10.0 * log10( fRefValue1/ (double)ssd2) : 99.99 );
		   psnr[2] = (ssd3 ? 10.0 * log10( fRefValue1/ (double)ssd3) : 99.99 );

		   //	printf("%d,%d,%d,\n",ssd1,ssd2,ssd3);
		   //printf("%8.4lf %8.4lf %8.4lf\n",psnr[0],psnr[1],psnr[2]);
}
